/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  config.c: configuration file parsing
  
  ***************************************/


#include <sys/stat.h>
#include <fcntl.h>
#include <netdb.h>
#include <pwd.h>
#include <grp.h>
#include <syslog.h>

#include "common.h"
#include "configutil.h"

void usage(int exitval);

struct acl_item {
	enum {ALLOW, DENY} action;
	struct in_addr src, smsk;
	struct in_addr dst, dmsk;
	u_int16_t ntp_ports[20][2];/*FIXME when I can be arsed to do it dynamically*/
};

struct options {
	char *config_file;
	char *logfile;
	char *pidfile;

/* Local Security stuff */
	struct sockaddr_in listen_address; 
	char *device;	    /* To bind to */
	char *user;	    /* To drop privelidges to */
	char *group;
	uid_t uid;
	gid_t gid;
	char *chroot;
	int maxforks;

	struct acl_item *acls;
	int acl_num;

/* Port ranges to use */
	u_int16_t hippasv,loppasv; 
	u_int16_t hipactv,lopactv;
	u_int16_t hipcont,lopcont;

/* Behaviour */
	struct timeval timeout;
	int same_address;
	int bounce_defend;

/* Optional extras */
	int transdata;
	int ntp;
	int apconv;

	char * cachemod;
	char *cacheopts[10];
};

void set_defaults(struct options *conf);
int parse_file(struct options * conf);
int parse_line(char *buf, struct options * conf);

int store_acl(struct options * conf);
int store_cacheopt(struct options *conf);
char * get_token(const char *s, const char *delim);

#ifdef DEBUG
void print_config(struct options *conf);
#endif

static struct options config;
static int chrooted=FALSE;

/* ------------------------------------------------------------- **
** Functions to read / parse the config file.
** ------------------------------------------------------------- */
int process_cmdline(int argc, char *argv[])
{
	int opt;

	set_defaults(&config);

	while((opt=getopt(argc,argv, "hf:")) !=-1) {
		switch(opt) {
		case 'f':
			config.config_file=strdup(optarg);
			break;
		case 'h':
			usage(0);
		case ':':
			fprintf(stderr, "Option -%c needs an argument\n", optopt);
			usage(-1);
		case '?':
			fprintf(stderr, "Unknown option -%c\n", optopt);
			usage(-1);
		}
	}
	if (optind!=argc) return(-1);
	return(0);
}

void usage(int exitval)
{	
	fprintf(stderr,"Usage frox [-f config_file] [-h]\n");
}

void set_defaults(struct options *conf)
{
	memset(conf, 0, sizeof(*conf));

	conf->loppasv=conf->lopactv=conf->lopcont=40000;
	conf->hippasv=conf->hipactv=conf->hipcont=50000;
	conf->timeout.tv_sec=3600;
	conf->listen_address.sin_addr.s_addr = htonl(INADDR_ANY);
	conf->listen_address.sin_family =  AF_INET;

	conf->config_file=conf->logfile=conf->pidfile=conf->device=
		conf->user=conf->group=conf->chroot=NULL;
	conf->acls = NULL;
	conf->cachemod = conf->cacheopts[0] = NULL;

	conf->bounce_defend=TRUE;
}

void config_free(struct options *conf)
{
	if(conf->acls) free(conf->acls);

	if(conf->config_file) free(conf->config_file);
	if(conf->logfile) free(conf->logfile);
	if(conf->pidfile) free(conf->pidfile);
	if(conf->device) free(conf->device);
	if(conf->user) free(conf->user);
	if(conf->group) free(conf->group);
	if(conf->chroot) free(conf->chroot);
}

int read_config(void)
{
       return(parse_file(&config));
}

int parse_file(struct options * conf)
{
	FILE *fp;
	char buf[BUF_LEN], buf2[BUF_LEN];
	int i;
	
	if(conf->config_file==NULL) fp=fopen(CONFIG_FILE,"r");
	else fp=fopen(conf->config_file,"r");
	if ( fp == NULL ) return(-1);

	do{
		if(fgets(buf,BUF_LEN,fp)==NULL) break;
		strcpy(buf2, buf);

		i=parse_line(buf2, conf);
		if(i==-1){
			fprintf(stderr, "Unrecognised argument in line:\n  %s",
				buf);
			return(-1);
		}
		if(i==-2){
			fprintf(stderr, "Unrecognised option in line:\n  %s",
				buf);
			return(-1);
		}
		
	}while(!feof(fp));

	fclose(fp);
	
	if(conf->chroot){ /*So that filename stripping works properly*/
		if(conf->chroot[strlen(conf->chroot)-1] == '/')
			conf->chroot[strlen(conf->chroot)-1] = 0;
	}
	
	/* We need to find out the uid / gid now. We must chroot
	** before changing priveleges, but if we wait until after
	** chroot we cannot parse /etc/passwd. 
	*/

	if(conf->group){
		struct group* grp;
		if( (grp=getgrnam(conf->group)) ==NULL){
			fprintf(stderr, "Unable to find group %s\n",
				conf->group);
			return(-3);
		}
		conf->gid=grp->gr_gid;
	}

	if(conf->user){
		struct passwd* pw;
		if( (pw=getpwnam(conf->user)) == NULL){
			fprintf(stderr, "Unable to find user %s\n",conf->user);
			return(-2);
		}
		conf->uid=pw->pw_uid;
	}
	
#if DEBUG
	print_config(conf);
#endif

	return(0);
}

/* ------------------------------------------------------------- **
** Returns: 0 Sucess/comment/blank line
**         -1 Invalid argument to option
**         -2 Unrecognised option.
** ------------------------------------------------------------- */
int parse_line(char *buf, struct options * conf)
{
	char *p;

	/* 
	** Remove leading whitespace. Skip comments and empty lines.
	*/
	for(p=buf;*p==' ' || *p=='\t';p++);
	if(*p=='#') return(0);
	memmove(buf, p, strlen(p) +1);
	if ((p = strchr(buf, '\n')) != NULL) *p = '\0';
	if(*buf=='\0' || *buf=='\n' ) return(0);
	p=get_token(buf, " \t=\n");
	
	if(!strcasecmp(p, "LogFile")) return(store_string(&conf->logfile));
	if(!strcasecmp(p, "PidFile")) return(store_string(&conf->pidfile));

	if(!strcasecmp(p, "Listen")) 
		return(store_address(&conf->listen_address.sin_addr));
	if(!strcasecmp(p, "Port"))
		return(store_port(&conf->listen_address.sin_port));
	if(!strcasecmp(p, "BindToDevice")) return(store_string(&conf->device));
	if(!strcasecmp(p, "User")) return(store_string(&conf->user));
	if(!strcasecmp(p, "Group")) return(store_string(&conf->group));
	if(!strcasecmp(p, "Chroot")) return(store_string(&conf->chroot));
	if(!strcasecmp(p, "MaxForks")) return(store_int(&conf->maxforks));

	if(!strcasecmp(p, "Timeout")) 
		return(store_int((int*) &conf->timeout.tv_sec));
	if(!strcasecmp(p, "PassivePorts"))
		return(store_portrange(&conf->loppasv, &conf->hippasv));
	if(!strcasecmp(p, "ActivePorts"))
		return(store_portrange(&conf->lopactv, &conf->hipactv));
	if(!strcasecmp(p, "ControlPorts"))
		return(store_portrange(&conf->lopcont, &conf->hipcont));
	if(!strcasecmp(p, "SameAddress"))
		return(store_bool(&conf->same_address));

	if(!strcasecmp(p, "ACL")) return(store_acl(conf));

	if(!strcasecmp(p, "TransparentData"))
		return(store_bool(&conf->transdata));
	if(!strcasecmp(p, "DoNTP")) return(store_bool(&conf->ntp));
	if(!strcasecmp(p, "APConv")) return(store_bool(&conf->apconv));
	if(!strcasecmp(p, "BounceDefend"))
		return(store_bool(&conf->bounce_defend));

#ifdef USE_CACHE
	if(!strcasecmp(p, "CacheModule"))
		return(store_string(&conf->cachemod));
	if(!strcasecmp(p, "Cache"))
		return(store_cacheopt(conf));
#endif

	return(-2);
}	


/* ------------------------------------------------------------- **
** Reread the configfile. A bit messy - maybe we should limit the
** things that can be changed on the fly?
** ------------------------------------------------------------- */
int reread_config(int *listening)
{
	int i;
	struct options newconfig;
	
	write_log(INFO, "Recieved SIGHUP - rereading config file");

	set_defaults(&newconfig);
	newconfig.config_file=strdup(config.config_file);
	
	i=parse_file(&newconfig);
	if(i==-1) return(-1);
	if(i==-3 && !chrooted) 
		write_log(ERROR, "Unable to determine UID/GID to drop to.");
	if(chrooted) newconfig.chroot=strdup(config.chroot);
	
	i = (config.listen_address.sin_port==newconfig.listen_address.sin_port &&
	     config.listen_address.sin_addr.s_addr==newconfig.listen_address.sin_addr.s_addr &&
	     !strcmp(config.device, newconfig.device));

	config_free(&config);
	config=newconfig;
	
	if(!chrooted) config_chroot();  /*No point re chrooting*/
	else strip_filename(config.chroot, config.logfile);

	if(!i){ /*Let's not close any sockets we don't have to*/
		int newfd;
		
		newfd=listen_on_socket(&config.listen_address, CTRL);
		if(newfd!=-1){
			close(*listening);
			*listening=newfd;
		} else {
			write_log(ERROR,"Ooops - can't change socket.");
		}
		config_bindtodevice(*listening);
	}

	config_openlog();
	config_droppriv();
	
	write_log(INFO, "Config file successfully reread");

#if DEBUG
	print_config(&config);
#endif

	return(0);
}

/* ------------------------------------------------------------- **
** Functions which implement config file settings.
** ------------------------------------------------------------- */
int config_bindtodevice(int fd)
{
	if(!config.device) return(0);
	if(setsockopt(fd, SOL_SOCKET, SO_BINDTODEVICE, 
		      (void *)config.device, (socklen_t) strlen(config.device)+1) != 0){
		perror("Binding to device");
		return(-1);
	}
	write_log(INFO, "Bound to device %s",config.device);
	return(0);
}
	
int config_chroot()
{
	if(!config.chroot) return(0);
	if( chroot(config.chroot)!=0 ||
	    chdir("/") !=0){
		free(config.chroot);
		config.chroot=NULL;
		return(-1);
	}
	write_log(INFO, "Chrooted to %s", config.chroot);
	chrooted=TRUE;
	
	/*Rewrite our filenames so they still work*/
	strip_filename(config.chroot, config.pidfile);
	strip_filename(config.chroot, config.logfile);
	strip_filename(config.chroot, config.config_file);
	
	debug("Chrooted:\n");
	debug2("  Now pidfile is %s\n", config.pidfile);
	debug2("  Now logfile is %s\n", config.logfile);
	return(0);
}

int config_droppriv()
{
	if(config.gid!=0){
		setgid(config.gid);
		setgid(config.gid);
	}
	if(config.uid!=0){
		if(config.transdata && sufd==0){ /*Kernel 2.2 we need root later*/
			seteuid(config.uid);
			write_log(INFO, "Temporarily dropped priveldges");
		} else {
			setuid(config.uid);
			setuid(config.uid);
			write_log(INFO, "Dropped priveledges");
		}
	}
	return(0);
}

int config_openlog()
{
	if(config.logfile){
		int tmpfd;
		tmpfd=open(config.logfile, O_APPEND|O_CREAT|O_WRONLY, S_IRUSR|S_IWUSR);
		if(tmpfd==-1 || dup2(tmpfd,2)==-1){
			fprintf(stderr, "Unable to open logfile %s\n",config.logfile);
			return(-1);
		}
	}
	openlog("frox", LOG_PID|LOG_CONS|LOG_NDELAY, LOG_DAEMON);
	return(0);
}

int config_cacheinit()
{
	if(config.cachemod)
		return(cache_geninit(config.cachemod, config.cacheopts, config.chroot));
	else return(0);
}

/* ------------------------------------------------------------- **
** Functions to return config file options.
** ------------------------------------------------------------- */
struct sockaddr_in config_getaddress()
{
	return(config.listen_address);
}

int config_loport(socketuse use)
{
	int ret=40000;
	
	switch(use) {
	case PASV:
		ret=config.loppasv;
		break;
	case ACTV:
		ret=config.lopactv;
		break;
	case CTRL:
		ret=config.lopcont;
		break;
	}
	
	return(ret);
}

int config_hiport(socketuse use)
{
	int ret=50000;
 
	switch(use) {
	case PASV:
		ret=config.hippasv;
		break;
	case ACTV:
		ret=config.hipactv;
		break;
	case CTRL:
		ret=config.hipcont;
		break;
	}

	return(ret);
}	
	
int config_maxforks()
{
	return(config.maxforks);
}

const char * config_pidfile()
{
	return(config.pidfile);
}

struct timeval config_timeout(void)
{
	return(config.timeout);
}

int config_apconv(void)
{
	return(config.apconv);
}

int config_ntp(void)
{
	return(config.ntp);
}

int config_usesameaddress()
{
	return(config.same_address);
}

int config_bouncedefend()
{
	return(config.bounce_defend);
}

int config_usecache()
{
	return(config.cachemod!=NULL);
}

int config_transdata()
{
#ifndef TRANS_DATA
	return(FALSE);
#else
	return(config.transdata);
#endif
}

void config_transdatafail(void)
{
	config.transdata=FALSE;
}

int config_connectionok(struct sockaddr_in *src, struct sockaddr_in *dst)
{
	struct acl_item *p;
	u_int16_t port;
	int i, j;

	port=ntohs(dst->sin_port);
	p=config.acls;
	for(i=0;i<config.acl_num; i++){
		if( (src->sin_addr.s_addr & p->smsk.s_addr) == p->src.s_addr &&
		    (dst->sin_addr.s_addr & p->dmsk.s_addr) == p->dst.s_addr){
			for(j=0;j<20 && p->ntp_ports[j][0]!=0;j++)
				if(port >= p->ntp_ports[j][0] &&
				   port <= p->ntp_ports[j][1])
					return(p->action==ALLOW);
		}
		p++;
	}
	return(FALSE);
}

int config_pasvok(struct sockaddr_in *addr)
{
	if(addr->sin_addr.s_addr==0) return(FALSE);
	
	if(config_usesameaddress()){
		if(addr->sin_addr.s_addr != 
		   info->server_control.address.sin_addr.s_addr){
			return(FALSE);
		}
	}
	return(TRUE);
}

int config_portok(struct sockaddr_in *addr)
{
	if(addr->sin_addr.s_addr==0) return(FALSE);

	if(config_usesameaddress())
		if(addr->sin_addr.s_addr !=
		   info->client_control.address.sin_addr.s_addr)
			return(FALSE);
	
	if(config_bouncedefend())
		if(ntohs(addr->sin_port)<1024) return(FALSE);
	
	return(TRUE);
}

/* ------------------------------------------------------------- **
** Alter absolute pathnames to still work after chroot.
** ------------------------------------------------------------- */
void strip_filename(const char *chroot, char *filename)
{
	if(filename==NULL) return;
	if (chroot==NULL) return;
	if(strncmp(chroot, filename, strlen(chroot))) *filename=0;
	else memmove(filename, filename+strlen(chroot),
		     strlen(filename) - strlen(chroot) + 1);
}

/* ------------------------------------------------------------- **
** Helper functions for config file parsing.
** ------------------------------------------------------------- */
int store_address(struct in_addr *addr)
{
	char *p;

	if((p=get_token(NULL, " \t:/-\n")) ==NULL) return(-1);

	if(strlen(p)==1 && *p=='*') {
		addr->s_addr=INADDR_ANY;
		return(0);
	}
	if(*p < '0' || *p >'9' ) {
		struct hostent *hostinfo;

		hostinfo = gethostbyname(p);
		if(hostinfo == NULL){
			fprintf(stderr, "Unable to resolve host %s\n", p);
			return(-1);
		}
		*addr = *((struct in_addr *) *hostinfo->h_addr_list);
	}
	else {
		if(inet_aton(p, addr)==0) return(-1);
	}
	return(0);
}

int store_port(u_int16_t *port)
{
	char *p;

	if((p=get_token(NULL, " \t\n")) ==NULL) return(-1);
	*port=strtol(p, NULL, 10);
	*port=htons(*port);
	return(0);
}


int store_addressport(struct sockaddr_in *addr)
{
	addr->sin_family=AF_INET;
	if(store_address(&addr->sin_addr)==-1) return(-1);
	return(store_port(&addr->sin_port));
}


int store_addressrange(struct in_addr *addr, struct in_addr *msk)
{
	char *p;
	int i;

	if(store_address(addr)==-1) return(-1);
	if(*get_token(NULL, NULL)=='/') {
		p=get_token(NULL, " \t-\n");
		if(strchr(p, '.')==NULL) { /* x.x.x.x/yy */
			i=atoi(p);
			msk->s_addr=htonl( 0xFFFFFFFF << (32-i));
			return(0);
		}
		else if(inet_aton(p, msk)==0) return(-1);
	}
	else { /*No mask given*/
		if(addr->s_addr==INADDR_ANY) msk->s_addr=INADDR_ANY;
		else msk->s_addr=INADDR_BROADCAST; 
	}
	return(0);
}

int store_portrange(u_int16_t *lo, u_int16_t *hi)
{
	char *p;

	if((p=get_token(NULL, " \t-\n")) ==NULL) return(-1);
	*lo=strtol(p, NULL, 10);
	if((p=get_token(NULL, " \t\n")) ==NULL) return(-1);
	*hi=strtol(p, NULL, 10);

	return(0);
}

int store_int(int *in)
{
	char *p;

	if((p=get_token(NULL, " \t\n")) ==NULL) return(-1);
	*in=strtol(p, NULL, 10);
	return(0);
}

int store_bool(int *in)
{
	char *p;

	if((p=get_token(NULL, " \t\n")) ==NULL) return(-1);

	if(!strcasecmp(p, "yes") ) *in=TRUE;
	else if(!strcasecmp(p, "no")) *in=FALSE;
	else return(-1);
	return(0);
}

int store_string(char **buf)
{
	char *p;

	if(*buf!=NULL) free(*buf);
	if((p=get_token(NULL, " \t\n")) ==NULL) return(-1);
	*buf=malloc(strlen(p)+1);
	if(*buf==NULL) quit(MALLOC);
	strcpy(*buf, p);
	return(0);
}

int store_cacheopt(struct options *conf)
{
	char *p, **pp;

	p=get_token(NULL, "");
	for(pp=conf->cacheopts; *pp!=NULL; pp++);
	*pp=strdup(p);
	*(pp+1)=NULL;
	return(0);
}

int store_acl(struct options *conf)
{
	int i;
	u_int16_t (*ports)[2];
	char *p;

	conf->acls=realloc(conf->acls, 
			       ++conf->acl_num * sizeof(struct acl_item));
	p=get_token(NULL, " \t\n");
	if(!strcasecmp(p, "Allow")) 
		(conf->acls+conf->acl_num-1)->action=ALLOW;
	else if(!strcasecmp(p, "Deny"))
		(conf->acls+conf->acl_num-1)->action=DENY;
	else return(-1);
	
	if(store_addressrange(&conf->acls[conf->acl_num-1].src,
			      &conf->acls[conf->acl_num-1].smsk)==-1)
		return(-1);
	if(store_addressrange(&conf->acls[conf->acl_num-1].dst,
			      &conf->acls[conf->acl_num-1].dmsk)==-1)
		return(-1);

	/*Now store the port range.*/
	ports=conf->acls[conf->acl_num-1].ntp_ports;
	for(i=0;i<20;i++){
		p=get_token(NULL, " \t,-\n");
		if(p==NULL) {
			if(i==0) { /*No ports specified - match all*/
				ports[i][0] = 1;
				ports[i][1] = 0xFFFF;
			}
			else ports[i][0] = ports[i][1] = 0;
			return(0);
		}
		if(strlen(p)==1 && *p=='*') {
			ports[i][0] = 1;
			ports[i][1] = 0xFFFF;
			return(0);
		}			
		ports[i][0] = atoi(p);
		if(*get_token(NULL, NULL)=='-') {
			p=get_token(NULL, " \t,-\n");
			if(p==NULL) ports[i][1]=0xFFFF;
			else ports[i][1]=atoi(p);
		} else ports[i][1] = ports[i][0];
	}
	fprintf(stderr, "Too many ports in port range\n");
	return(-1);
}

/* ------------------------------------------------------------- **
** A more useful replacement to strtok(3). Identical except: 
**  o The first argument is not modified. 
**  o If the delimiter string is empty we return the whole remaining 
**    string but don't discard it. Subsequent calls can therefore 
**    continue to process it. 
**  o If delim==NULL we return the character which delimited the 
**    last token we returned. 
** ------------------------------------------------------------- */
char * get_token(const char *s, const char *delim)
{
	static char ret[256], buf[256]={0}, sep=0;
	int i;

	if(s!=NULL){
		if(strlen(s)>255) return(NULL);
		strcpy(buf, s);
	}
	if(delim==NULL){
		ret[0]=sep;
		return(ret);
	}
	if(*delim=='\0') {
		strcpy(ret, buf);
		return(ret);
	}

	i=strcspn(buf, delim);
	sep=buf[i];
	strncpy(ret, buf, i);
	ret[i]=0;
	i+=strspn(buf+i, delim); /*Strip out consecutive delimiters*/
	memmove(buf, buf+i, strlen(buf+i)+1);
	if(*ret==0) return(NULL);
	else return(ret);
}

#ifdef DEBUG
void print_config(struct options *conf)
{
	int i, j;

	fprintf(stderr, "\n\nconfig_file = %s\n",conf->config_file);
	fprintf(stderr, "logfile = %s\n",conf->logfile);
	fprintf(stderr, "pidfile = %s\n\n",conf->pidfile);

	fprintf(stderr, "listen_address = %s:%d\n", 
		inet_ntoa(conf->listen_address.sin_addr), 
		ntohs(conf->listen_address.sin_port));
	fprintf(stderr, "device = %s\n", conf->device);
	fprintf(stderr, "user = %s(%d)\n",conf->user, conf->uid);
	fprintf(stderr, "group = %s(%d)\n",conf->group, conf->gid);
	fprintf(stderr, "chroot = %s\n", conf->chroot);
	fprintf(stderr, "maxforks = %d\n\n", conf->maxforks);

	for(i=0;i<config.acl_num;i++){
		fprintf(stderr, "ACL: %s", inet_ntoa(conf->acls[i].src));
		fprintf(stderr, "/%s", inet_ntoa(conf->acls[i].smsk));
		fprintf(stderr, " --> %s", inet_ntoa(conf->acls[i].dst));
		fprintf(stderr, "/%s  ", inet_ntoa(conf->acls[i].dmsk));
		fprintf(stderr, ": %s ", 
			conf->acls[i].action==DENY ? "DENY" : "ALLOW");
		for(j=0; j<20 && conf->acls[i].ntp_ports[j][0]!=0;j++){
			fprintf(stderr, "%d-%d,", 
				conf->acls[i].ntp_ports[j][0],
				conf->acls[i].ntp_ports[j][1]);
		}
		fprintf(stderr, "\n");
			
	}

	fprintf(stderr, "\nPortranges = %d-%d, %d-%d, %d-%d\n\n",
		conf->loppasv,conf->hippasv,
		conf->lopactv,conf->hipactv,
		conf->lopcont,conf->hipcont);

	fprintf(stderr, "same_address = %s\n", conf->same_address?"On":"Off");
	fprintf(stderr, "bounce_defend = %s\n",conf->bounce_defend?"On":"Off");
	fprintf(stderr, "transdata = %s\n", conf->transdata?"On":"Off");
	fprintf(stderr, "ntp = %s\n", conf->ntp?"On":"Off");
	fprintf(stderr, "apconv = %s\n", conf->apconv?"On":"Off");

	fprintf(stderr, "Cache opts are:\n");
	for(i=0;i<10 && conf->cacheopts[i]!=0; i++)
		fprintf(stderr, "  %s\n", conf->cacheopts[i]);
}
#endif
