/***************************************

    This is part of frox: A simple transparent FTP proxy
    Copyright (C) 2000 James Hollingshead

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

  cache.c -- Stuff for requesting files through an external cache (eg. squid)


  ***************************************/
#include <fcntl.h>
#include "common.h"
#include "configutil.h"
#include "cache.h"
#include "control.h"

extern char * get_token(const char *s, const char *delim);

char* read_httpline(char *inc, int *len);
static enum {STARTING, SUCCEEDING, SUCCESS, FAILURE, NONE} cache_status=NONE;
static int min_size=0, offset;
static struct sockaddr_in http_proxy;

int s_geninit(char **cacheopts, char *chroot)
{
	char *p, **pp;

	for(pp=cacheopts; *pp!=NULL; pp++){
		p=get_token(*pp, " \t=");
		if(!strcasecmp(p, "HTTPProxy")) store_addressport(&http_proxy);
		else if(!strcasecmp(p, "MinCacheSize")) 
			store_int(&min_size);
		else return(-1);
	}
	fprintf(stderr, "%s   %d\n", inet_ntoa(http_proxy.sin_addr), min_size);	
	return(0);
}

/* ------------------------------------------------------------- **
** If we are to retrieve through squid then return a fd. If not, or on
** error, return -1. 
** ------------------------------------------------------------- */
int s_retr_start(const char *host, const char *file, const char *mdtm, int size, int offst)
{
	char rangestr[MAX_LINE_LEN];
	char buf[BUF_LEN];
	int i, squidfd;

	offset=offst;
	if(min_size>size){
		cache_status=NONE;
		debug("File too small for cache - retrieving directly\n");
		return(-1);
	}

	/*Set up the HTTP command*/
	cache_status=STARTING;

	if(offset>0) sprintf(rangestr, "Range: bytes=%d-\r\n", offset);
	else rangestr[0]=0;
	i=snprintf(buf, BUF_LEN,
		   "GET ftp://%s%s HTTP/1.0\r\n"
		   "Host: %s\r\n"
		   "User-Agent: %s\r\n"
		   "%s\r\n",
		   host, file,
		   host,
		   "Frox/0.5",
		   rangestr );

	if(i<0 || i>BUF_LEN) quit(CLIENT_RUBBISH);

	debug2("HTTP string = %s\n", buf);

	if( (squidfd=connect_to_socket(http_proxy, PASV))==-1){
		write_log(ERROR, "Unable to contact HTTP proxy. Retrieving directly");
		cache_status=NONE;
		return(-1);
	}

	/*Send the message to squid. We could block for a while here,
	 * but no-one should be sending us anything important...*/
	fcntl(squidfd, F_SETFL, O_NONBLOCK);
	write_all(squidfd, buf);

	send_message(150,"Starting transfer");

	/*And set everything up as if the connection was normal*/

	return(squidfd);
}

/* ------------------------------------------------------------- **
** Called whenever we select on incoming data from server. Strip out
** HTTP lines, and if the request failed send a failure message and
** make sure we continue to ignore everything.
** ------------------------------------------------------------- */
void s_inc_data(char *inc, int *len)
{
	char *buf, *p;

	if(cache_status==NONE) return;
	if(cache_status==FAILURE){ /*Discard message body*/
		debug("Discarding HTTP body after failure.\n");
		*len=0;
		return;
	}
	if(cache_status==SUCCESS) return;
	
	/*Parse HTTP reply. FIXME - parse it properly!*/
	do{
		buf=read_httpline(inc, len);
		if(buf==NULL) return; /*Incomplete line*/
		
		if( !strncmp(buf,"HTTP", 4)){ /*Status Line*/
			for(p=strchr(buf,' '); *p < '0' || *p > '9'; p++); 
			if(*p=='2' && (offset==0 || *(p+2)=='6') ) cache_status=SUCCEEDING;
			else{ /*Failure - discard rest of message.*/
				*len=0;
				cache_status=FAILURE;
				return;
			}
		}
		else if(strlen(buf)==0) 
			cache_status= (cache_status==SUCCEEDING) ? SUCCESS : FAILURE;

	}while(cache_status!=SUCCESS && cache_status!=FAILURE);

	return;
}

void s_retr_end(void)
{
	if(cache_status==NONE) return;
	if(cache_status==SUCCESS) send_message(226, "Transfer complete");
	else send_message(426, "Unable to transfer file");
	cache_status=NONE;
	return;
}

/* ------------------------------------------------------------- **
** Remove one line from client_data buffer. return it (without
** terminating \r\n), or NULL if no complete line exists.
** ------------------------------------------------------------- */
char* read_httpline(char *inc, int *len)
{
	static char buf[MAX_LINE_LEN];
	int i;

	for(i=0; i<*len; i++) 
		if(inc[i]=='\r') break;
	if(inc[i]!='\r') return(NULL); /*Incomplete line*/	

	inc[i]=0;
	strncpy(buf, inc, MAX_LINE_LEN-1);
	buf[MAX_LINE_LEN-1]=0;

	i+=2;
	memmove(inc, inc+i, *len-i+1);
	*len-=i;
	
	debug2("HTTP: %s\n",buf);
	return(buf);
}
